#! /usr/bin/env python
"""
gap_candidate.py

Created by Lucas Swanson
Copyright (c) 2012 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

# import custom modules
from utils.general import AddChr, StrJoin
from utils.messages import ErrMsg
from common.coord_pair import CoordPairCls, BetweenCoords, CutOrExtendBlocks
from gene_overlap import GenesOverlapped, NearbyGenes
from alignment_functions import FixAlign, CalcAlignOverlap, AlignString

# CONSTANTS
MAX_LOCAL_DIST = 5000
PERFECT_IDENTITY = 100.0

#Class to handle gap alignments from a single contig
class GapCandidateCls: #{
  # align1 = original alignment of the contig to the genome
  # align2 = pseudoalignment of the gap sequence to the rest of contig
  # contig id
  # contig_rep = total # of bases aligned in the two alignments
  # contig_rep_fraction = fraction of total contig represented by the two alignments
  # alignment topology:
  #   for now, just 'gap-duplication'
  # multi_mapped = 1 if align1 is multi-mapped

  def __init__(self, align1, gap_with_dup, add_gene_annotation, ctg_seq): #{
    # Read in the original alignment of the contig to the genome.
    self.align1 = FixAlign(align1)
    # Read in the pseudoalignment of the gap sequence to the rest of the contig.
    self.align2 = MatchedGapPseudoAlignmentCls(align1, gap_with_dup)
    # Read in the attributes of the contig
    self.contig = align1.query
    self.contig_size = int(align1.query_len)
    # Calculate the contig representation and fraction
    #   (use the gap size as the 'overlap' value).
    (self.contig_rep, self.contig_rep_fraction) = self.CalcCtgRepresentation(
        self.align1, self.align2, gap_with_dup.Span())
    # Set the topology to "gap-duplication" by default.
    self.topology = "gap-duplication"
    # Count how many of the alignments are spliced.
    self.spliced = 0
    if (self.align1.spliced): #{
      self.spliced += 1
    #} end if
    if (self.align2.spliced): #{
      self.spliced += 1
    #} end if
    # Assume that align1 is not multi-mapped.
    self.multi_mapped = 0
    # Keep track of whether we should be adding gene annotation.
    self.add_gene_annotation = add_gene_annotation
    self.ctg_seq = ctg_seq
  #} end def

  # Calculate the amount of the contig that is represented by the combination
  # of the two alignments and the fraction of the total contig length that
  # this represents
  def CalcCtgRepresentation(self, align1, align2, olap): #{
    # TODO use qcoords CoordPairCls object
    covered_by_align1 = (align1.qend - align1.qstart) + 1
    covered_by_align2 = (align2.qend - align2.qstart) + 1
    overlap = max(0, olap)
    total_covered = covered_by_align1 + covered_by_align2 - overlap
    fraction_covered = float(total_covered) / float(align1.query_len)
    return (total_covered, fraction_covered)
  #} end def

  def CheckMultiMapping(self): #{
    if (self.align1.multi_mapped): #{
      self.multi_mapped += 1
    #} end if
  #} end def

  def BreakpointsString(self): #{
    # Determine the "breakpoint_from" target coordinate
    breakpoint_from = self.align2.tcoords.min
    # Determine the "breakpoint_to" target coordinate
    breakpoint_to = self.align2.tcoords.max
    return "%s:%i(%s)-%s:%i(%s)" % (
      AddChr(self.align1.target), breakpoint_from, 'down',
      AddChr(self.align2.target), breakpoint_to, 'up')
  #} end def

  def Details(self): #{
    ctg_fields = self.contig.split(" ")
    ctg = "%s(%ibp)" % (ctg_fields[0], self.contig_size)
    # TODO use coords object
    target = "%s:%i-%i,%s:%i-%i" % (
      AddChr(self.align1.target), self.align1.tstart, self.align1.tend,
      AddChr(self.align2.target), self.align2.tstart, self.align2.tend)
    query = "%i-%i,%i-%i" % (
      self.align1.qstart, self.align1.qend,
      self.align2.qstart, self.align2.qend)
    # TODO use coords object, do not need to use float here
    align_len1 = float(self.align1.qend) - float(self.align1.qstart) + 1
    if (hasattr(self.align2, 'gap_coords')): #{
      align_len1 -= self.align2.gap_coords.Span()
    #} end if
    align_len2 = float(self.align2.qend) - float(self.align2.qstart) + 1
    align_frac1 = align_len1/float(self.contig_size)
    align_frac2 = align_len2/float(self.contig_size)
    align_metrics = ",".join([
      "I1:%.2f"  % float(self.align1.identity),
      "I2:%.2f"  % float(self.align2.identity),
      "AF1:%.2f" % align_frac1,
      "AF2:%.2f" % align_frac2])
    blocks1 = ",".join(["-".join(map(str, block)) for
      block in self.align1.blocks])
    blocks2 = ",".join(["-".join(map(str, block)) for
      block in self.align2.blocks])
    blocks = ";".join([blocks1, blocks2])
    #olap_genes1 = GenesOverlapped(self.align1, self.add_gene_annotation)
    #olap_genes2 = GenesOverlapped(self.align2, self.add_gene_annotation)
    #olap_genes = ";".join([olap_genes1, olap_genes2])
    #nearby_genes1 = NearbyGenes(self.align1, self.add_gene_annotation)
    #nearby_genes2 = NearbyGenes(self.align2, self.add_gene_annotation)
    #nearby_genes = ";".join([nearby_genes1, nearby_genes2])
    #gap_metrics = ",".join((
    #  "CO:%ibp(%.3f)" % (self.ctg_overlap, self.ctg_overlap_fraction),
    #  "CR:%.3f"       % self.contig_rep_fraction,
    #  "BD:%s"         % self.breakpoint_distance,
    #  "TD:%s"         % self.target_distance,
    #  "QG:%i"         % self.qgaps,
    #  "SP:%i"         % self.spliced,
    #  "GF:%s"         % self.gapped_event_found,
    #  "MM:%i"         % self.multi_mapped,
    #))
    detail_string = " ".join((
      "CTG:%s"      % ctg,
      "TOPOLOGY:%s" % self.topology,
      "TARGET:%s"   % target,
      "CONTIG:%s"   % query,
      "BREAKPOINTS:%s" % self.BreakpointsString(),
      "%s"          % align_metrics,
      "BLOCKS:%s"   % blocks,
      "GENES:%s"    % self.main_genes_string(),
      "NEARBY:%s"   % self.near_genes_string(),
      "META:%s"     % self.meta_data_string(),
      'EVENT_SEQ:%s' % self.dup_seq(),
      'INS_SEQ:%s' % self.ins_seq(),
    ))
    return detail_string
  #} end def

  def main_genes_string(self): #{
    return GenesOverlapped(self.align1, self.add_gene_annotation)
  #} end def

  def near_genes_string(self): #{
    return NearbyGenes(self.align1, self.add_gene_annotation)
  #} end def

  def meta_data_string(self): #{
    return ",".join((
      'CS:{}'.format(self.align1.query_strand),
      'ES:{}'.format(self.align2.query_strand),
      'GAP:{}'.format(self.align2.gap_coords),
      'NO_GAP:{}'.format(StrJoin(';', self.no_gap_pair())),
      'DUP:{}'.format(self.align2.dup_coords),
      'DIST:{}'.format(self.dup_dist()),
      'TRIM:N/A;N/A',
      'FORM:N/A',
      'MM:0'.format(self.multi_mapped),
    ))
  #} end def

  def no_gap_pair(self): #{
    """Determine the genomic coordinates of the two pieces flanking the gap.

    """
    before_gap_coords = self._no_gap_coords(self.align2.before_blocks)
    after_gap_coords = self._no_gap_coords(self.align2.after_blocks)
    ErrMsg('No-gap Coordinates: {}'.format(
      StrJoin(';', (before_gap_coords, after_gap_coords))))
    return (before_gap_coords, after_gap_coords)
  #} end def

  def _no_gap_coords(self, block_list): #{
    if not block_list: #{
      return "N/A-N/A"
    #} end if
    gap_start = min((block.min for block in block_list))
    gap_end = max((block.max for block in block_list))
    # Swap start and end if aligned to the negative strand.
    if '-' == self.align1.query_strand:
      gap_start, gap_end = gap_end, gap_start
    return CoordPairCls(gap_start, gap_end)
  #} end def

  def dup_dist(self):
    """Return the distance between the two copies of the duplicated sequence.

    """
    return len(self.ins_seq())
  #} end def

  def dup_seq(self):
    """Return the sequence that is duplicated."""
    dup_seq = self.ctg_seq[
        self.align2.dup_coords.min-1:self.align2.dup_coords.max]
    ErrMsg('Duplicated sequence: {}'.format(dup_seq))
    return dup_seq
  #} end def

  def ins_seq(self):
    """Return the sequence (if any) between the copies of the duplicated
    sequence.

    """
    between_coords = BetweenCoords(
        self.align2.qcoords, self.align2.dup_coords)
    ins_seq = self.ctg_seq[
        between_coords.min-1:between_coords.max]
    return ins_seq
  #} end def
#} end class


class MatchedGapPseudoAlignmentCls: #{
  def __init__(self, main_align, gap_with_dup): #{
    # qcoords represents the portion of the gap involved in the duplication.
    self.qcoords = _get_gap_qcoords(gap_with_dup)
    ErrMsg('Gap qcoords: {}'.format(self.qcoords))
    self.qstart, self.qend = self.qcoords.start, self.qcoords.end
    # TODO figure out blocks (genomic coordinates)
    self.blocks = self.get_dup_blocks(main_align, gap_with_dup)
    ErrMsg('Dup blocks: {}'.format(StrJoin(',', self.blocks)))
    self.spliced = (1 < len(self.blocks))
    self.target = main_align.target
    self.tstart = min((block.min for block in self.blocks))
    self.tend = max((block.max for block in self.blocks))
    # Swap start and end if aligned to the negative strand.
    if '-' == main_align.query_strand:
      self.tstart, self.tend = self.tend, self.tstart
    self.tcoords = CoordPairCls(self.tstart, self.tend)
    ErrMsg('Target Coordinates: {}'.format(self.tcoords))
    # NOTE: Identity assumes aligned copy has no mismatches with genome.
    self.identity = PERFECT_IDENTITY
    self.query_strand = main_align.query_strand
    # gap_coords represents the contig coordinates of the full gap.
    self.gap_coords = gap_with_dup.Copy()
    # dup_coords represents the portion of the original alignment
    # involved in the duplication.
    self.dup_coords = _get_gap_dup_coords(gap_with_dup)
  #} end def

  def get_dup_blocks(self, main_align, gap_with_dup): #{
    ErrMsg('Getting duplication blocks...\nMain alignment:\n{}\n'
        'Query blocks: {}\nTarget blocks: {}\nGap: {}'.format(
            AlignString(main_align),
            StrJoin(',', main_align.qblock_coords),
            StrJoin(',', main_align.tblock_coords),
            gap_with_dup))
    # zip query and target blocks, traverse to split zipped lists
    # at gap position, using gap contig coordinates
    block_list_pair = zip(main_align.qblock_coords, main_align.tblock_coords)
    if '-' == main_align.query_strand: #{
      block_list_pair.reverse()
    #} end if
    self.before_blocks = list()
    self.after_blocks = list()
    for index, (qblock, tblock) in enumerate(block_list_pair): #{
      if (qblock.Span() != tblock.Span()): #{
        raise GapCandidateError(
            'error getting duplicate blocks for alignment, query and target '
            'block lengths disagree:\n  {}'.format(AlignString(main_align)))
      #} end if
      if qblock.min > gap_with_dup.max: #{
        ErrMsg('Gap occurs before block {}'.format(index))
        self.after_blocks.append(tblock)
      else:
        self.before_blocks.append(tblock)
      #} end if
    #} end for
    if gap_with_dup.match_is_before_gap: #{
      from_left = ('-' == main_align.query_strand)
      blocks_to_cut = self.before_blocks[:]
    else:
      from_left = ('+' == main_align.query_strand)
      blocks_to_cut = self.after_blocks[:]
    #} end if
    if '-' == main_align.query_strand: #{
      blocks_to_cut.reverse()
    #} end if
    # Make an initial cut, a little too long, to account for the offset
    cut_blocks = CutOrExtendBlocks(
        blocks_to_cut, gap_with_dup.match_len + gap_with_dup.toffset, from_left)
    # Cut again from the opposite end, to get rid of the extra from the offset
    return CutOrExtendBlocks(cut_blocks, gap_with_dup.match_len, not from_left)
  #} end def
#} end class


def _get_gap_qcoords(gap_with_dup): #{
  # Return contig coordinates for the portion of the gap involved in the
  # duplication.
  ErrMsg('Getting gap duplication query coordinates...')
  if gap_with_dup.match_is_before_gap: #{
    ErrMsg('  match is before gap')
    qend = gap_with_dup.max
    ErrMsg('  gap max: {}, qend: {}'.format(gap_with_dup.max, qend))
    return CoordPairCls(end=qend, span=gap_with_dup.match_len)
  else:
    ErrMsg('  match is after gap')
    qstart = gap_with_dup.min
    ErrMsg('  gap min: {}, qstart: {}'.format(gap_with_dup.min, qstart))
    return CoordPairCls(qstart, span=gap_with_dup.match_len)
  #} end if
#} end def


def _get_gap_dup_coords(gap_with_dup): #{
  # Return contig coordinates for the portion of the original alignment
  # involved in the duplication.
  ErrMsg('Getting gap duplication duplicated region coordinates...')
  if gap_with_dup.match_is_before_gap: #{
    ErrMsg('  match is before gap')
    qend = gap_with_dup.min - (gap_with_dup.toffset+1)
    ErrMsg('  gap min: {}, qend: {}'.format(gap_with_dup.min, qend))
    return CoordPairCls(end=qend, span=gap_with_dup.match_len)
  else:
    ErrMsg('  match is after gap')
    qstart = gap_with_dup.max + (gap_with_dup.toffset+1)
    ErrMsg('  gap max: {}, qstart: {}'.format(gap_with_dup.max, qstart))
    return CoordPairCls(qstart, span=gap_with_dup.match_len)
  #} end if
#} end def


def _get_dup_blocks(main_align, gap_with_dup): #{
  ErrMsg('Getting duplication blocks...\nMain alignment:\n{}\n'
      'Query blocks: {}\nTarget blocks: {}\nGap: {}'.format(
          AlignString(main_align),
          StrJoin(',', main_align.qblock_coords),
          StrJoin(',', main_align.tblock_coords),
          gap_with_dup))
  # zip query and target blocks, traverse to split zipped lists
  # at gap position, using gap contig coordinates
  block_list_pair = zip(main_align.qblock_coords, main_align.tblock_coords)
  if '-' == main_align.query_strand: #{
    block_list_pair.reverse()
  #} end if
  gap_with_dup.before_blocks = list()
  gap_with_dup.after_blocks = list()
  for index, (qblock, tblock) in enumerate(block_list_pair): #{
    if (qblock.Span() != tblock.Span()): #{
      raise GapCandidateError(
          'error getting duplicate blocks for alignment, query and target '
          'block lengths disagree:\n  {}'.format(AlignString(main_align)))
    #} end if
    if qblock.min > gap_with_dup.max: #{
      ErrMsg('Gap occurs before block {}'.format(index))
      gap_with_dup.after_blocks.append(tblock)
    else:
      gap_with_dup.before_blocks.append(tblock)
    #} end if
  #} end for
  if gap_with_dup.match_is_before_gap: #{
    from_left = ('-' == main_align.query_strand)
    blocks_to_cut = gap_with_dup.before_blocks[:]
  else:
    from_left = ('+' == main_align.query_strand)
    blocks_to_cut = gap_with_dup.after_blocks[:]
  #} end if
  if '-' == main_align.query_strand: #{
    blocks_to_cut.reverse()
  #} end if
  # Make an initial cut, a little too long, to account for the offset
  cut_blocks = CutOrExtendBlocks(
      blocks_to_cut, gap_with_dup.match_len + gap_with_dup.toffset, from_left)
  # Cut again from the opposite end, to get rid of the extra from the offset
  return CutOrExtendBlocks(cut_blocks, gap_with_dup.match_len, not from_left)
#} end def


def _get_no_gap_pair(gap_with_dup, strand): #{
  before_gap_start = min((block.min for block in gap_with_dup.before_blocks))
  before_gap_end = max((block.max for block in gap_with_dup.before_blocks))
  after_gap_start = min((block.min for block in gap_with_dup.after_blocks))
  after_gap_end = max((block.max for block in gap_with_dup.after_blocks))
  # Swap start and end if aligned to the negative strand.
  if '-' == strand:
    before_gap_start, before_gap_end = before_gap_end, before_gap_start
    after_gap_start, after_gap_end = after_gap_end, after_gap_start
  before_gap_coords = CoordPairCls(before_gap_start, before_gap_end)
  after_gap_coords = CoordPairCls(after_gap_start, after_gap_end)
  return (before_gap_coords, after_gap_coords)
#} end def


#### EXCEPTION CLASSES ####
class GapCandidateError(Exception): #{
  def __init__(self, msg): #{
    self.msg = msg
  def __str__(self): #{
    return repr(self.msg)
#} end class
