#! /usr/bin/env python
"""
aligned_coord_pair.py

Created by Lucas Swanson
Copyright (c) 2012 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

# import custom modules
from utils.error import MyError
from common.coord_pair import CoordPairCls, strand_string

class AlignedCoordPairCls: #{
  def __init__(self, qname, qcoords, tname, tcoords): #{
    # qcoords and tcoords should be CoordPairCls objects
    # Query coordinates should always be considered "positive strand".
    if not qcoords.pos_strand:
      raise AlignedCoordPairError(
          'Cannot create AlignedCoordPairCls with query coordinates: {0}. '
          'Query coordinates should always be considered '
          '"positive strand".'.format(qcoords))
    # Query and target coordinates should span the same distance.
    if qcoords.Span() != tcoords.Span():
      raise AlignedCoordPairError(
          'Cannot create AlignedCoordPairCls with query coordinates: '
          '{0} ({1}), and target coordinates: {2} ({3}). Coordinates must '
          'have equal spans.'.format(
            qcoords, qcoords.Span(), tcoords, tcoords.Span()))
    self.qname = qname
    self.qcoords = qcoords
    self.tname = tname
    self.tcoords = tcoords
  #} end def

  def pos_strand(self):
    # Strand is determined based on target coordinates.
    return self.tcoords.pos_strand
  #} end def

  def __str__(self):
    return '{0}:{1}={2}:{3}'.format(
        self.qname, self.qcoords, self.tname, self.tcoords)
#} end class

class AlignedCoordPairListCls(AlignedCoordPairCls): #{
  def __init__(self, qname, tname): #{
    AlignedCoordPairCls.__init__(
        self, qname, CoordPairCls(), tname, CoordPairCls())
    self.coord_pairs = list()
  #} end def

  def add_coord_pair(self, qcoords, tcoords): #{
    new_pair = AlignedCoordPairCls(self.qname, qcoords, self.tname, tcoords)
    # If coord_pairs have already been saved in this list, make sure that
    # the new pair has the same strand.
    if self.coord_pairs and tcoords.pos_strand != self.pos_strand(): #{
      raise AlignedCoordPairError(
          'Cannot add target coordinates: {0}. Strand does not match strand '
          'of list: {1}'.format(tcoords, strand_string(self.pos_strand())))
    #} end if
    # If coord_pairs have already been saved in this list, make sure that
    # new pairs are added in order.
    if self.coord_pairs: #{
      if self.qcoords.end >= qcoords.start: #{
        raise AlignedCoordPairError(
            'Cannot add query coordinates: {0}, since they do not come after '
            'the query coordinates already in the list: {1}'.format(
              qcoords, self.qcoords))
      #} end if
      if self.pos_strand: #{
        if self.tcoords.end >= tcoords.start: #{
          raise AlignedCoordPairError(
              'Cannot add target coordinates: {0}, since they do not come after '
              'the target coordinates already in the list: {1}'.format(
                tcoords, self.tcoords))
        #} end if
      else:
        if self.tcoords.end <= tcoords.start: #{
          raise AlignedCoordPairError(
              'Cannot add target coordinates: {0}, since they do not come before '
              'the target coordinates already in the list: {1}'.format(
                tcoords, self.tcoords))
        #} end if
      #} end if
    #} end if
    self.qcoords.Union(qcoords)
    self.tcoords.Union(tcoords)
    self.coord_pairs.append(new_pair)
  #} end def
#} end class

# TODO: Add function to get 'gaps' in AlignedCoordPairCls (either query, or
#       target).

#### EXCEPTION CLASSES ####
class AlignedCoordPairError(MyError): #{
  pass
#} end class
