#! /usr/bin/env python
"""
psl.py

Created by Readman Chiu
Edited by Lucas Swanson
Copyright (c) 2012 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

# TODO: Fix long lines.

# import standard modules
from __future__ import division
import re, sys

# import custom modules
from utils.general import ReverseComplement, IsInteger
from utils.messages import ErrMsg, DebugMsg, LogMsg
from utils.files_paths import FileBoxCls
from parsers.alignment import (ContigAlignmentParser, Alignment,
    ContigAlignmentParserError, calc_score)

# constants
# PSL_HEADER_LENGTH does not include any blank lines.
PSL_HEADER_LENGTH = 4
COL_MATCH = 0
COL_MISMATCH = 1
COL_REPMATCH = 2
COL_NCOUNT = 3
COL_QNUMINSERT = 4
COL_QBASEINSERT = 5
COL_TNUMINSERT = 6
COL_TBASEINSERT = 7
COL_QSTRAND = 8
COL_QNAME = 9
COL_QLEN = 10
COL_QSTART = 11
COL_QEND = 12
COL_TNAME = 13
COL_TLEN = 14
COL_TSTART = 15
COL_TEND = 16
COL_BLOCK_COUNT = 17
COL_BLOCK_SIZES = 18
COL_QSTARTS = 19
COL_TSTARTS = 20
PSL = 'psl'

def parse_str_field(field_value):
    return field_value

def parse_int_field(field_value):
    # Integer fields that do not require base-change.
    return int(field_value)

def parse_base_change_field(field_value):
    # Integer fields that require base-change.
    return int(field_value)+1

class FieldDefinitionCls:
    def __init__(self, name, parse_function=None, is_minimum=False, do_parse=True):
        self.name = name
        # 'parse' is the function used to parse the field (e.g. convert to
        #   int or float, add 1, etc.).
        if parse_function is None:
            self.parse = parse_int_field
        else:
            self.parse = parse_function
        # 'is_minimum' is true iff the field should be included in the
        #   alignment objects when the minimum flag is set to true.
        self.is_minimum = is_minimum
        # 'do_parse' is False for fields that are not used directly.
        self.do_parse = do_parse

FIELD_DEFINITTIONS = [
    FieldDefinitionCls('match'),
    FieldDefinitionCls('mismatch', is_minimum=True),
    FieldDefinitionCls('repmatch'),
    FieldDefinitionCls('ncount'),
    FieldDefinitionCls('qnuminsert'),
    FieldDefinitionCls('qbaseinsert'),
    FieldDefinitionCls('tnuminsert'),
    FieldDefinitionCls('tbaseinsert'),
    FieldDefinitionCls(
        'query_strand', parse_function=parse_str_field, is_minimum=True),
    FieldDefinitionCls(
        'query', parse_function=parse_str_field, is_minimum=True),
    FieldDefinitionCls('query_len'),
    FieldDefinitionCls('qstart', parse_function=parse_base_change_field),
    FieldDefinitionCls('qend'),
    FieldDefinitionCls(
        'target', parse_function=parse_str_field, is_minimum=True),
    FieldDefinitionCls('target_len'),
    FieldDefinitionCls('tstart', parse_function=parse_base_change_field),
    FieldDefinitionCls('tend'),
    FieldDefinitionCls('block_count'),
    FieldDefinitionCls('block_sizes', do_parse=False),
    FieldDefinitionCls('qstarts', do_parse=False),
    FieldDefinitionCls('tstarts', do_parse=False),
]
# None of these fields is optional.
MIN_NUM_FIELDS = len(FIELD_DEFINITTIONS)


# TODO: Change most log messages in member functions to extreme debug messages.
class PSLContigAlignmentParser(ContigAlignmentParser):
    def __init__(
            self, psl_path, filters, noblocks, minimum, noline, refseq,
            log_info=None):
        ContigAlignmentParser.__init__(
            self, psl_path, filters, noblocks, minimum, noline, refseq,
            log_info=log_info)
        # prev_line is used to check for duplicated lines, which are
        # sometimes seen in malformed PSL files.
        self.prev_line = None

    def is_header(self, line):
        """Return true if line is a header line (does not start with an
        integer).

        """
        return not IsInteger(line[0])

    def parse_header(self):
        """Parse the header (if any) of the PSL-formatted file with the given
        path. After this function is called, self.curr_line will hold the
        first line of the body of that file (the first alignment line).

        A PSL header is a certain number of lines that do not start with an
        integer, but some files might not have a header.

        May raise StopIteration.

        """
        self.curr_line = self.file.next()
        # If the first line starts with an integer, then assume that the file
        # does not have any header.
        if not self.is_header(self.curr_line):
            return
        for i in range(PSL_HEADER_LENGTH):
            if not self.is_header(self.curr_line):
                raise PSLParserError(
                    'PSL header lines should not start with an integer!\n'
                    '  {0}'.format(self.curr_line))
            self.curr_line = self.file.next()

    def process_body_line(self):
        # A duplicated line is likely a sign of an error, so raise an
        # exception if the current line is a duplicate of the previous line.
        if self.curr_line == self.prev_line:
            raise PSLParserError(
                'Duplicate line encountered:\n{0}'.format(self.curr_line))
        self.prev_line = self.curr_line
        # Use the base class member function to continue processing the
        # current line.
        ContigAlignmentParser.process_body_line(self)

    def align_is_incomplete(self, align_cols):
        # Note that value of MIN_NUM_FIELDS is format-specific.
        return len(align_cols) < MIN_NUM_FIELDS

    def query_is_unaligned(self, align_cols):
        """Unaligned queries are never present in PSL alignment files."""
        return False

    def align_fails_ungapped_filter(self, align):
        """An ungapped alignment in PSL format should have only a single
        block, and the number of blocks is stored explicitly.

        """
        if ('ungapped' in self.filters and
            1 < len(align.query_blocks)):
            return True
        return False

    def create_align(self, record):
        """Create and return a PSL alignment object from the provided record,
        which should be a string read from the PSL-formatted alignment
        file.

        """
        field_values = record.split("\t")
        align = PSLAlignment()
        for field_def, field_value in zip(FIELD_DEFINITTIONS, field_values):
            if field_def.do_parse and (
               field_def.is_minimum or not self.minimum):
                setattr(align, field_def.name, field_def.parse(field_value))
        if not self.minimum:
            try:
                align.SetScore()
            except ValueError, e:
                raise PSLParserError(
                    'Error while setting the score for alignment of {0}:\n'
                    '{1}.'.format(align.query, e))
            align.set_num_bases_aligned()
            align.SetIdentity()
        if not self.noline:
            setattr(align, 'psl_str', record)
        if not self.noblocks:
            align.blocks = get_blocks(
                field_values[COL_TSTARTS], field_values[COL_BLOCK_SIZES])
            align.query_blocks = get_blocks(
                field_values[COL_QSTARTS], field_values[COL_BLOCK_SIZES],
                strand=field_values[COL_QSTRAND], qsize=field_values[COL_QLEN])
        # Splice sites and orientation.
        if align.blocks and self.refseq: #{
            get_splice_sites(align, self.refseq)
        # PSL alignments are ranked by alignment score, as defined by UCSC.
        align.rank_by = align.score
        align.raw_record = record
        return align


# TODO: The only parameters ever used are psl_path, filters, and log_info!
# TODO: Which filters are actually used? Maybe we really can do one alignment
#       at a time (treat each individually), rather than having to load all
#       alignments for a contig at once.
# TODO: Remove genome parameter.
def parse(psl_path, filters=None, noblocks=False, noline=False, minimum=False,
          splice_motif_file=None, refseq=None, genome=None, log_info=None): #{
    # If no filters are provided, alignments are processed one at a time.
    # Otherwise all alignments for a contig are processed at once.
    # TODO: This should be a generator, that yields one alignment at a time,
    #       rather than loading them all into memory at once!!
    #       ...might need to parse all alignments for a contig, then yield
    #       those one at a time.
    parser = PSLContigAlignmentParser(
            psl_path, filters, noblocks, minimum, noline, refseq,
            log_info=log_info)
    parser.parse_header()
    ErrMsg('parser.curr_line =\n  {0}'.format(parser.curr_line))
    # TODO: Add parser.parse_body function.
    # Infinite loop, rely on StopIteration to break out when file is complete.
    while True:
        if parser.is_header(parser.curr_line):
            raise ContigAlignmentParserError(
                'Header line found in body of alignments file:\n'
                '  {0}'.format(parser.curr_line))
        try:
            parser.process_body_line()
        except StopIteration:
            break
    parser.file.close()
    # Process the alignments for the 'last' contig in the alignment file.
    parser.process_alignment_group()
    return parser.aligns


# TODO: Move this to PSLAlignment.
# TODO: Once this is a member function, strand and qsize will be accessible
#       as data members, so the parameters can be replaced with
#       are_query_blocks boolean parameter.
# TODO: Rename to set_blocks.
def get_blocks(starts, block_sizes, strand=None, qsize=None): #{
    # TODO: Use coord_pairs data member instead of blocks.
    blocks = []
    # TODO: Add psl_block_info_list function.
    # TODO: Use map(int, ...) in psl_block_info_list.
    starts = starts.rstrip(',').split(',')
    blk_sizes = block_sizes.rstrip(',').split(',')
    # TODO: qsize is necessary here! So raise an exception if it is not provided?
    #       qsize is not given/needed for 'target' blocks.
    if strand and qsize and strand == '-': #{
        for i in range(len(starts)): #{
            end = int(qsize)-int(starts[i])
            start = end - int(blk_sizes[i]) + 1
            blocks.append([end, start])
    else:
        for i in range(len(blk_sizes)): #{
            block = [int(starts[i])+1, int(starts[i])+int(blk_sizes[i])]
            blocks.append(block)
        #blocks.sort(lambda x,y: x[0]-y[0])
    return blocks


def get_splice_sites(align, refseq): #{
    if refseq.has_key(align.target): #{
        for i in range(len(align.blocks)-1): #{
            if align.blocks[i+1][0] - align.blocks[i][1] < 5: #{
                ss = "NA"
            else:
                intron = refseq[align.target].seq.tostring()[align.blocks[i][1]:align.blocks[i+1][0]-1]
                ss = intron[:2].lower() + intron[-2:].lower()
            align.splice_sites.append(ss)
            #print i, ss
    #print "splice",align.splice_sites, len(align.splice_sites)


# TODO: Replace blocks with target_blocks.
class PSLAlignment(Alignment):
    def __init__(self):
        Alignment.__init__(self, PSL)

#### EXCEPTION CLASSES ####
class PSLParserError(ContigAlignmentParserError): #{
  """Exception raised for errors encountered while parsing PSL format
  contig-to-genome alignments.

  """
  pass
#} end class

if __name__ == '__main__': #{
    parse(sys.argv[1])
