#! /usr/bin/env python
"""
sam.py

Created by Readman Chiu
Edited by Lucas Swanson
Copyright (c) 2012 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

# TODO: Fix long lines.

# import standard modules
from __future__ import division
import re, sys

# import custom modules
from utils.general import StrJoin, IsInteger
from utils.messages import ErrMsg, TestMsg, LogMsg
from utils.common_constants import FIRST, LAST, POS_STRAND, NEG_STRAND
from parsers.alignment import (
    ContigAlignmentParser, Alignment, ContigAlignmentParserError)
from parsers.tokenizer import GetFieldAndValue

# constants
# TODO: Move these constants!
I_OP = 0
I_LEN = 1
FIRST_TAG_COL = 11
MIN_MATCH_LENGTH = 4
HCOL_TNAME = 1
HCOL_TLEN = 2
COL_QNAME = 0
COL_FLAG = 1
COL_RNAME = 2
COL_POS = 3
COL_MAPQ = 4
COL_CIGAR = 5
COL_RNEXT = 6
COL_PNEXT = 7
COL_TLEN = 8
COL_SEQ = 9
COL_QUAL = 10
SAM = 'sam'
HEADER_START = '@'
TARGET_START = '{0}SQ'.format(HEADER_START)
FN_TARGET = 'SN'
FN_TARGET_LEN = 'LN'
# A target value of '*' indicates that the query is actually unaligned.
UNALIGNED_TARGET_MARKER = '*'
# TODO: Add CIGAR operation constants and use them instead of string literals.
SECOND = 1
SECOND_TO_LAST = -2

# NOTE: fields is not used anywhere. What is its purpose?
FIELD_NAMES = (
    'qname flag rname pos mapq cigar rnext pnext tlen seq qual').split()
MIN_NUM_FIELDS = len(FIELD_NAMES)

cigar_pattern_global = re.compile('\d+[A-Z]')
cigar_pattern_local = re.compile('(\d+)([A-Z])')
del_pattern = re.compile(r'\^[ACGT]+')
# NOTE: choice of 100 as size threshold is fairly arbitrary
large_del_pattern = re.compile(r'\^[ACGT]{100,}')
misalign_pattern = re.compile(r'([ACGT]+)?\^?[ACGT]+([0-3]([ACGT]+)?\^?[ACGT]+)+')
md_patt = re.compile(r'(\d+|\^?[A-Z]+)')
poor_flank_patterns = {
    'before': re.compile(r'[ACGT]*\^?[ACGT]+[0-9]$'),
    'after':  re.compile(r'^[0-9]\^?[ACGT]+')
}


# TODO: Change most log messages in member functions to extreme debug messages.
class SAMContigAlignmentParser(ContigAlignmentParser):
    def __init__(
            self, sam_path, filters, noblocks, minimum, noline, refseq,
            log_info=None):
        ContigAlignmentParser.__init__(
            self, sam_path, filters, noblocks, minimum, noline, refseq,
            log_info=log_info)
        self.target_lens = None

    def is_header(self, line):
        """Return true if line is a header line (starts with '@')."""
        return line.startswith(HEADER_START)

    def parse_header(self):
        """Parse the header of the SAM-formatted file with the given
        path. After this function is called, self.curr_line will hold the
        first line of the body of that file (the first alignment line).

        May raise StopIteration.

        """
        self.target_lens = dict()
        self.body_line_encountered = False
        for self.curr_line in self.file:
            if self.is_header(self.curr_line):
                if not self.curr_line.startswith(TARGET_START):
                    continue
                hfields = self.curr_line.split('\t')
                # Check field names to ensure they are 'SN' and 'LN',
                # respectively.
                field_name, target = GetFieldAndValue(hfields[HCOL_TNAME])
                if FN_TARGET != field_name:
                    raise SAMParserError(
                        'Target name field in SAM file header (column {0}) '
                        'should be "{1}", not "{2}".'.format(
                            HCOL_TNAME, FN_TARGET, field_name))
                field_name, target_len = GetFieldAndValue(hfields[HCOL_TLEN])
                if FN_TARGET_LEN != field_name:
                    raise SAMParserError(
                        'Target length field in SAM file header (column {0}) '
                        'should be "{1}", not "{2}".'.format(
                            HCOL_TLEN, FN_TARGET_LEN, field_name))
                self.target_lens[target] = int(target_len)
            else:
                # Assume we are done the header as soon as we get to a line
                # that does not start with '@'.
                self.body_line_encountered = True
                break
        # If no target lengths were found, raise an error.
        if not self.target_lens:
            raise SAMParserError(
                'No target lengths found in SAM file: {0}'.format(
                    self.file.path))

    def align_is_incomplete(self, align_cols):
        # Note that value of MIN_NUM_FIELDS is format-specific.
        return len(align_cols) < MIN_NUM_FIELDS

    def query_is_unaligned(self, align_cols):
        """SAM-formatted alignments show 'unaligned' status with a special
        target value.

        """
        return align_cols[COL_RNAME] == UNALIGNED_TARGET_MARKER

    def align_fails_ungapped_filter(self, align):
        """An ungapped alignment in SAM format should have a CIGAR string
        with a single 'M' operation, of any length.

        """
        if ('ungapped' in self.filters and
            re.match(r'^\d+M$', align.cigar) is None):
            return True
        return False

    def create_align(self, record):
        """Create and return a SAM alignment object from the provided record,
        which should be a string read from the SAM-formatted alignment
        file.

        """
        if self.target_lens is None:
          raise SAMParserError(
              'target_lens cannot be None in SAMContigAlignmentParser '
              'create_align.')
        field_values = record.split("\t")
        align = SAMAlignment()
        align.query = field_values[COL_QNAME]
        TestMsg("Creating alignment for %s" % align.query)
        align.target = field_values[COL_RNAME]
        if align.target == UNALIGNED_TARGET_MARKER:
            # TODO: Return alignment object as is, replace outside checks for
            #       alignment being None with checks for target?
            return None
        align.target_len = self.target_lens.get(align.target)
        if align.target_len is None:
            raise SAMParserError(
                'Could not get target length for alignment with target: %s. '
                'Known targets: %s' % (align.target,
                StrJoin(", ", sorted(self.target_lens.keys()))))
        # TODO: Use value from field_values as 'check_tstart', set actual tstart based
        #       on blocks, using AlignedCoordPairListCls functionality.
        align.tstart = int(field_values[COL_POS])
        align.mapping_quality = int(field_values[COL_MAPQ])
        # TODO: Use this value here as 'check_pos_strand', set actual pos_strand
        #       based on blocks, using AlignedCoordPairListCls functionality?
        if int(field_values[COL_FLAG]) & 16:
            align.query_strand = NEG_STRAND
        else:
            align.query_strand = POS_STRAND
        # TODO: Add parse_cigar_and_MD SAMAlignment member function.
        align.raw_cigar = field_values[COL_CIGAR]
        align.tags = ParseTags(field_values)
        # TODO: blocks_before_fix should be based on coord_pairs Alignment data member.
        blocks_before_fix = align.get_blocks_from_cigar()[0]
        # TODO: tend_before_fix will be easier to get after blocks_before_fix
        #       is coord_pairs.
        tend_before_fix = blocks_before_fix[LAST][LAST]
        # At this point, target coordinates should be in non-stranded order.
        # TODO: Should be able to use pos_strand here.
        if tend_before_fix < align.tstart: #{
          raise SAMParserError("Error parsing alignment for contig %s: "
            "tend %i < tstart %i" % (align.query, tend_before_fix, align.tstart))
        #} end if
        try:
            # compare CIGAR and MD size
            if align.cigar_and_md_size_differ():
                raise SAMParserError(
                    'CIGAR and MD strings imply different lengths for contig '
                    '{0}\n  CIGAR: {1}\n  MD: {2}'.format(
                        align.query, align.cigar, align.tags.get("MD")))
            # TODO: Add adjust_tstart and adjust_tend alignment data members.
            # TODO: Use adjust_tcoords CoordPairCls for adjust_tstart and adjust_tend.
            (adjust_tstart,
                adjust_tend) = align.handle_possible_misalignments()
        except Exception, e:
            raise SAMParserError("Error parsing alignment for contig %s: %s" %
                (align.query, e))
        TestMsg("Adjust tstart: %i Adjust tend: %i" % (adjust_tstart, adjust_tend))
        # Save the 'pre-adjustment' tstart coordinate (for use in re-sorting
        # alignments to return 'original order').
        # TODO: Do we need to save otstart if we get rid of the sorting/original
        #       order stuff?
        align.otstart = align.tstart
        align.tstart += adjust_tstart
        # Note that process_cigar_and_md sets align.match and align.mismatch.
        if not align.process_cigar_and_md():
            # TODO: Raise error!
            return None
        # Ensure that fixing CIGAR did not mess up target coordinate ordering
        # (should still pretend to be pos_strand).
        # TODO: Use 'pos_strand'.
        if align.tend < align.tstart: #{
          raise SAMParserError("Error parsing alignment for contig %s: "
            "tend %i < tstart %i" % (align.query, align.tend, align.tstart))
        #} end if
        # Ensure that the target end coordinate adjustment was applied correctly.
        if abs(tend_before_fix - align.tend) != adjust_tend: #{
          raise SAMParserError('Error parsing alignment for contig %s: fixing '
              'CIGAR altered tend (before: %i, after: %i)' %
              (align.query, tend_before_fix, align.tend))
        #} end if
        # Ensure that fixing CIGAR did not mess up mismatches.
        # TODO: Use Span member function.
        if align.match + align.mismatch > (align.qend - align.qstart) + 1: #{
          raise SAMParserError("Error parsing alignment for contig %s: fixing "
              "CIGAR messed up mismatches (match: %i, mismatch: %i, "
              "span: %i)" % (align.query, align.match, align.mismatch,
                (align.qend-align.qstart)-1))
        #} end if
        # SAM alignments are ranked by MAPQ.
        align.rank_by = align.mapping_quality
        align.raw_record = record
        return align


# TODO: The only parameters ever used are sam_path, filters, and log_info!
# TODO: Add 'sort_body' parameter, default to False (because it will mean we
#       need to load ALL alignments, rather than just the ones for the
#       current contig.)
# TODO: Which filters are actually used? Maybe we really can do one alignment
#       at a time (treat each individually), rather than having to load all
#       alignments for a contig at once. We use 'count' and 'identity'.
# TODO: Keep a list of contigs processed, to check for body being out of order.
#       Require body is 'sorted' by contig (contig order of alignments is the
#       same as the contig order of the input sequence files).
# TODO: Make these parameters identical to those in psl.py.
def parse(sam_path, filters=None, noblocks=False, minimum=False, noline=None,
          refseq=None, original_order=True, log_info=None):
    # If no filters are provided, alignments are processed one at a time.
    # Otherwise all alignments for a contig are processed at once.
    # TODO: This should be a generator, that yields one alignment at a time,
    #       rather than loading them all into memory at once!!
    #       ...might need to parse all alignments for a contig, then yield
    #       those one at a time.
    parser = SAMContigAlignmentParser(
            sam_path, filters, noblocks, minimum, noline, refseq,
            log_info=log_info)
    # TODO: Add general parse_alignment_file member function.
    # TODO: Directly return results of parse_alignment_file function.
    parser.parse_header()
    if not parser.body_line_encountered:
        print 'No alignments found!'
        return list()
    ErrMsg('parser.curr_line =\n  {0}'.format(parser.curr_line))
    # TODO: Add parser.parse_body function.
    # Infinite loop, rely on StopIteration to break out when file is complete.
    while True:
        if parser.is_header(parser.curr_line):
            raise ContigAlignmentParserError(
                'Header line found in body of alignments file:\n'
                '  {0}'.format(parser.curr_line))
        try:
            parser.process_body_line()
        except StopIteration:
            break
    parser.file.close()
    # Process the alignments for the 'last' contig in the alignment file.
    parser.process_alignment_group()
    return parser.aligns


# TODO: Move this to SAMAlignment?
def get_num_matches_from_MD(md_tag):
    if md_tag is None: #{
      return 0
    #} end if
    total_matches = 0
    matches = re.findall("\d+", md_tag.value)
    for m in matches:
        total_matches += int(m)
    return total_matches


# TODO: Move this to SAMAlignment?
def get_num_mismatches_from_MD(md_tag): #{
    TestMsg("Getting number of mismatches...")
    if md_tag is None: #{
      return 0
    #} end if
    mismatches = 0
    #for m in re.findall(r"[ACGT]", md_tag.value): #{
    #  mismatches += 1
    #} end for
    # adjust for deletions
    #for m in re.findall(r"^[ACGT]+", md_tag.value): #{
    #  mismatches -= (len(m) - 1)
    #} end for
    TestMsg("  MD: %s" % md_tag.value)
    for m in re.findall(r"\^?[ACGT]+", md_tag.value): #{
      #TestMsg("  md_piece: %s" % m)
      if not m.startswith("^"): #{
        mismatches += len(m)
      #} end if
    #} end for
    TestMsg("  found %i mismatches" % mismatches)
    return mismatches


# TODO: Move this to SAMAlignment?
def get_cigar_ops(cigar_string):
    ops_str = cigar_pattern_global.findall(cigar_string)
    ops = []
    for length_op in ops_str:
        m = cigar_pattern_local.match(length_op)
        length, op = m.group(1), m.group(2)
        ops.append([op, int(length)])
    return ops


# TODO: This should not be necessary, once using alignment object properly.
def query_align_bounds(cigar_string):
    """Extract qstart and qend."""
    ops = get_cigar_ops(cigar_string)
    qstart = 1
    i = 0
    while i < len(ops):
        op, length = ops[i]
        if op == 'M':
            break
        #soft or hard clipped
        elif op == 'S' or op == 'H':
            qstart += length
        i += 1
    qend = get_query_len(cigar_string)
    i = len(ops) - 1
    while i >= 0:
        op, length = ops[i]
        if op == 'M':
            break
        elif op == 'S' or op == 'H':
            qend -= length
        i -= 1
    return qstart, qend


# TODO: This should not be necessary, once using alignment object properly.
def get_query_len(cigar_string):
    ops = get_cigar_ops(cigar_string)
    lens = []
    for op, length in ops:
        if op != 'D' and op != 'N' and op != 'P':
            lens.append(int(length))
    return sum(lens)


class TagCls:
    def __init__(self, tag_string=None, value=None):
        if tag_string is not None:
            # TagCls should not be initialized with both tag_string and value.
            assert value is None
            self._parse_tag_string(tag_string)
        else:
            self.value = value

    def _parse_tag_string(self, tag_string):
        (self.id, self.ttype, self.value) = tag_string.split(':')
        if "i" == self.ttype:
            self.value = int(self.value)
        elif "f" == self.ttype:
            self.value = float(self.value)

    def __str__(self):
        return str(self.value)


def ParseTags(align_cols):
    tag_list = dict()
    for align_col in align_cols[FIRST_TAG_COL:]:
        new_tag = TagCls(tag_string=align_col)
        tag_list[new_tag.id] = new_tag
    return tag_list


def GetMDIndex(md_list, md_bases): #{
  remaining = md_bases
  for index, md in enumerate(md_list): #{
    if IsInteger(md): #{
      remaining -= int(md)
    elif "^" == md[0]:
      remaining -= (len(md)-1)
    else:
      remaining -= len(md)
    #} end if
    if 0 > remaining: #{
      return index
    #} end if
  #} end for
  raise SAMParserError("Cannot get MD index: too many bases. MD: %s; "
      "bases: %i" % (",",join(md_list), md_bases))
#} end def


def tokenize_md(md_string):
    md_tokens = md_patt.findall(md_string)
    for i,token in enumerate(md_tokens):
        if IsInteger(token):
            md_tokens[i] = int(token)
    return md_tokens


def get_lengths_from_cigar(cigar):
    cigar_qsize = cigar_tsize = 0
    if isinstance(cigar, str):
        cigar = get_cigar_ops(cigar)
    for (op, oplen) in cigar:
        if op in "HSIM":
            cigar_qsize += oplen
        if op in "NDM":
            cigar_tsize += oplen
    return (cigar_qsize, cigar_tsize)


def md_before_and_after_should_be_merged(md_dict):
    if (md_dict['before'] and md_dict['after'] and
        IsInteger(md_dict['before'][LAST]) and
        IsInteger(md_dict['after'][FIRST])):
        return True
    return False

def merge_md_before_and_after(md_dict):
    TestMsg('Merging before and after...')
    md_dict['before'][LAST] = (
        int(md_dict['before'][LAST]) + int(md_dict['after'][FIRST]))
    del md_dict['after'][FIRST]


# TODO: Replace blocks with target_blocks.
class SAMAlignment(Alignment):
    def __init__(self):
        Alignment.__init__(self, SAM)
        self.cigar = None
        self.md_tag = None

    def cigar_and_md_size_differ(self):
        size_from_md = self._get_size_from_md()
        size_from_cigar = self._get_size_from_cigar()
        TestMsg("CIGAR says MD length should be: %i, actual MD length: %i" %
            (size_from_cigar, size_from_md))
        if size_from_cigar != size_from_md:
            TestMsg("Skipping alignment with unmatched CIGAR and MD strings...")
            return True
        return False

    def handle_possible_misalignments(self):
        old_md = self.tags.get('MD')
        # Assume that there will not be any misalignments.
        self.cigar = self.raw_cigar
        if old_md is None:
            TestMsg('No MD string provided, cannot fix CIGAR and MD.')
            return (0, 0)
        TestMsg('Fixing CIGAR string: %s' % self.raw_cigar)
        TestMsg('And MD string of length: %i, start: %s' %
            (len(old_md.value), old_md.value[:150]))
        old_md.value = self._fix_large_deletions(old_md.value)
        self.misalignment = MisalignedCIGARAndMDCls(
            old_md.value, self.raw_cigar)
        if self.misalignment.md_match is None:
            self.cigar = self.raw_cigar
            self.tags['MD'] = old_md
            return (0, 0)
        self.was_misaligned = True
        # TODO: Add MisalignedCIGARAndMDCls.fix_md_and_cigar function to call the below.
        self.misalignment._setup_md_dict()
        self.misalignment._setup_cigar_dict()
        self.misalignment._fix_md()
        self.misalignment._fix_cigar()
        self.cigar = self.misalignment.fixed_cigar
        self.tags['MD'] = TagCls(value=self.misalignment.fixed_md)
        # TODO: Update values of data members directly, rather than returning
        #       values. Remember to check other returns.
        return (self.misalignment.tadjustments[FIRST],
                self.misalignment.tadjustments[LAST])

    def _get_size_from_md(self):
        if self.tags.get("MD") is None:
            raise SAMParserError(
                'No MD string provided, cannot calculate MD size.')
        size_from_md = 0
        md_list = tokenize_md(self.tags["MD"].value)
        # Add up sizes of items in MD tag.
        for md_item in md_list:
            if IsInteger(md_item):
                size_from_md += int(md_item)
            elif "^" == md_item[0]:
                size_from_md += len(md_item)-1
            else:
                size_from_md += len(md_item)
        return size_from_md

    def _get_size_from_cigar(self):
        size_from_cigar = 0
        if self.cigar is None:
            cigar_to_use = self.raw_cigar
        else:
            cigar_to_use = self.cigar
        cigar_list = get_cigar_ops(cigar_to_use)
        # count "D" and "M" CIGAR operations
        for (op, length) in cigar_list:
            if op in "DM":
                size_from_cigar += length
        return size_from_cigar

    def _fix_large_deletions(self, old_md_string):
        TestMsg('Fixing large deletions in CIGAR and MD strings...')
        # first check for missing carets (^)
        new_md_string = self._check_for_missing_carets_in_md(old_md_string)
        # TODO: Maybe should use CIGAR rather than MD to find large deletions?
        adjust = 0
        for m in large_del_pattern.finditer(new_md_string):
            deletion = LargeDeletionCls(m, adjust)
            #TestMsg('    Check: %s' %
            #        new_md_string[deletion.start: deletion.start+20])
            md_dict = dict({
                'before': tokenize_md(new_md_string[:deletion.start]),
                'after': tokenize_md(new_md_string[deletion.end:]),
            })
            #TestMsg('    Last-Before: %s, First-After: %s' % (
            #    md_dict['before'][LAST], md_dict['after'][FIRST]))
            #TestMsg('    B: %i, A: %i' % (len(md_dict['before']), len(md_dict['after'])))
            #TestMsg('    BEFORE: %s' % StrJoin('', md_dict['before']))
            if md_before_and_after_should_be_merged(md_dict):
                merge_md_before_and_after(md_dict)
            len_before = len(new_md_string)
            new_md_string = StrJoin('', md_dict['before'] + md_dict['after'])
            len_after = len(new_md_string)
            self.raw_cigar = deletion.update_cigar(self.raw_cigar)
            #TestMsg('    Fixed: %s, %s' % (new_md_string, self.raw_cigar))
            #TestMsg('    Fixed: length %i, start = %s..., %s' %
            #    (len(new_md_string), new_md_string[:150], self.raw_cigar))
            adjust += (len_before - len_after)
        #TestMsg('  Fixed: %s, %s' % (new_md_string, self.raw_cigar))
        TestMsg('  Fixed: start = %s..., %s' % (
            new_md_string[:150], self.raw_cigar))
        return new_md_string

    def _check_for_missing_carets_in_md(self, old_md_string):
        if self.raw_cigar.count("D") == old_md_string.count("^"):
            return old_md_string
        cigar_list = get_cigar_ops(self.raw_cigar)
        md_list = tokenize_md(old_md_string)
        md_bases = 0
        for (op, length) in cigar_list:
            if "M" == op:
                md_bases += length
            elif "D" == op:
                md_index = GetMDIndex(md_list, md_bases)
            if "^" != md_list[md_index][0]:
                TestMsg("  adding ^ to MD")
                md_list[md_index] = "^%s" % md_list[md_index]
            TestMsg("  MD piece: %s" % md_list[md_index])
            if (del_pattern.search(md_list[md_index]) is None or
                length != len(md_list[md_index])-1):
                raise SAMParserError("Error fixing deletions in MD tag: %s" %
                    md_list[md_index])
            md_bases += length
        return "".join(md_list)

    # TODO: Break this super long function up into smaller bits.
    def get_blocks_from_cigar(self):
        if self.cigar is None:
            cigar_to_use = self.raw_cigar
        else:
            cigar_to_use = self.cigar
        # TODO: Probably worthwhile to make cigar_ops data member.
        ops = get_cigar_ops(cigar_to_use)
        # TODO: Do we already have query_len at this point?
        query_len = get_query_len(cigar_to_use)
        # TODO: This line is completely unnecessary, just make sure tstart
        #       is already int.
        tstart = int(self.tstart)
        # TODO: Why not use self.qstart?
        if '+' == self.query_strand:
          qstart = 1
        else:
          qstart = query_len
        self.blocks = []
        self.query_blocks = []
        first_block = True
        for op, length in ops:
          if 'D' == op and first_block:
              # TODO: Raise an error.
              return None, None
          if 'S' == op or 'H' == op:
            if first_block:
              if '+' == self.query_strand:
                qstart = qstart + length
              else:
                qstart = qstart - length
              first_block = False
              continue
            # TODO: Add 'else' clause, for internal clipping?
          first_block = False
          if 'M' != op and not self.blocks:
              # TODO: Raise an error.
              return None, None
          # match
          if 'M' == op:
            tend = tstart + length - 1
            if '+' == self.query_strand:
              qend = qstart + length - 1
            else:
              qend = qstart - length + 1
            self.blocks.append([tstart, tend])
            self.query_blocks.append([qstart, qend])
          # intron, skipped reference or deletion in reference
          elif 'D' == op or 'N' == op:
            #intron
            if 'N' == op and 3 > length:
              continue
            #qend = qend
            tend = tstart + length - 1
          #insertion to reference
          elif 'I' == op:
            #tend = tend
            if '+' == self.query_strand:
              qend = qstart + length - 1
            else:
              qend = qstart - length + 1
          tstart = tend + 1
          if '+' == self.query_strand:
            qstart = qend + 1
          else:
            qstart = qend - 1
        return self.blocks, self.query_blocks

    # TODO: Rename this to something more descriptive.
    #       e.g. _count_matches_mismatches_from_MD
    def process_MD(self):
        #ErrMsg("In process_MD: %s" % self.tags.get('MD'))
        self.match = get_num_matches_from_MD(self.tags.get("MD"))
        self.mismatch = get_num_mismatches_from_MD(self.tags.get("MD"))

    def get_indels_from_CIGAR(self):
        ins = re.findall('([1-9]\d*)I', self.cigar)
        self.qnuminsert = len(ins)
        self.qbaseinsert = sum(int(i) for i in ins)
        dels = re.findall('([1-9]\d*)[D]', self.cigar)
        self.tnuminsert = len(dels)
        self.tbaseinsert = sum(int(i) for i in dels)
        skips = re.findall('([1-9]\d*)[N]', self.cigar)
        num_skips = len(skips)
        bases_skips = sum(int(i) for i in skips)
        return num_skips, bases_skips

    def process_cigar_and_md(self):
        self.query_len = get_query_len(self.cigar)
        # TODO: Do not need to use return value, values already set within
        #       member function.
        self.blocks, self.query_blocks = self.get_blocks_from_cigar()
        if self.blocks is None or self.query_blocks is None:
            # TODO: Raise error!
            return False
        # TODO: Replace 0 with FIRST and 1 with SECOND.
        self.qstart = min(self.query_blocks[FIRST][0],
                          self.query_blocks[FIRST][1],
                          self.query_blocks[LAST][0],
                          self.query_blocks[LAST][1])
        self.qend = max(self.query_blocks[FIRST][0],
                        self.query_blocks[FIRST][1],
                        self.query_blocks[LAST][0],
                        self.query_blocks[LAST][1])
        self.tend = self.blocks[LAST][1]
        # The following is all for the UCSC psl track.
        self.process_MD()
        num_skips, bases_skips = self.get_indels_from_CIGAR()
        # no negative value accepted
        if self.mismatch is not None:
            # TODO: Should raise error if mismatch is None or negative mismatch
            #       value is found.
            self.mismatch = max(0, self.mismatch)
        # this seems like a silly way to handle 'skips'
        # why not just treat N and D together in get_indels_from_CIGAR?
        if bases_skips is not None:
            self.tbaseinsert += bases_skips
        if num_skips is not None:
            self.tnuminsert += num_skips
        try:
            self.SetScore()
        except ValueError, e:
            raise SAMParserError(
                'Error while setting the score for alignment of {0}:\n'
                '{1}.'.format(self.query, e))
        self.set_num_bases_aligned()
        self.SetIdentity()
        return True

    def fix_match_for_double_miss(self, match, is_left_match):
        self.misalignment.fix_cigar_and_md_for_double_miss(
            match, is_left_match)
        self.cigar = self.misalignment.fixed_cigar
        self.tags['MD'].value = self.misalignment.fixed_md
        self.process_cigar_and_md()
        # TODO: Replace qend-qstart+1 with span().
        if self.match + self.mismatch > (self.qend - self.qstart) + 1:
            raise SAMParserError("Error parsing alignment for contig %s: fixing "
                "CIGAR messed up mismatches (match: %i, mismatch: %i, "
                "span: %i)" % (self.query, self.match, self.mismatch,
                    (self.qend-self.qstart)-1))
        # reset psl string
        self.psl_str = None


class MisalignedCIGARAndMDCls:
    def __init__(self, md_string, raw_cigar):
        TestMsg("  Constructing MisalignedCIGARAndMDCls class")
        self.md_string = md_string
        self.raw_cigar = raw_cigar
        self.md_match = None
        for m in misalign_pattern.finditer(md_string):
            TestMsg('    MD misalign: {}'.format(m.group()))
            if (self.md_match is None or
                len(self.md_match.group()) < len(m.group())):
                self.md_match = m
        self.cigar_qsize = self.cigar_tsize = 0
        self.md_dict = dict()
        self.cigar_dict = dict()
        self.clipping = [["S",0], ["S",0]]
        self.tadjustments = [0, 0]
        self.poor_flank = None

    def _setup_md_dict(self):
        TestMsg("  Setting up MD dictionary")
        str_dict = {"misalign": self.md_match.group()}
        TestMsg("    Found misalignment in MD: %s" % str_dict["misalign"])
        str_dict["before"] = self.md_string[:self.md_match.start()]
        str_dict["after"]  = self.md_string[self.md_match.end():]
        # grow mismatch if flanking pieces are also poor
        for flank in ["before", "after"]:
            # TODO: Rename PoorMDFlank to flank_has_poor_md.
            while self.PoorMDFlank(str_dict, flank):
                TestMsg("    Growing misalign into %s flank" % flank)
                self.GrowMismatch(str_dict, flank)
        # if before or after are a single digit from 1-3, move it to misalign
        if str_dict["before"] in ["1", "2", "3"]:
            str_dict["misalign"] = str_dict["before"] + str_dict["misalign"]
            str_dict["before"] = ""
        if str_dict["after"] in ["1", "2", "3"]:
            str_dict["misalign"] = str_dict["misalign"] + str_dict["after"]
            str_dict["after"] = ""
        TestMsg("    Adjusted misalignment MD: %s" % str_dict["misalign"])
        TestMsg("      Before: %s\n      After: %s" %
            (str_dict["before"], str_dict["after"]))
        for md_key in str_dict:
            self.md_dict[md_key] = tokenize_md(str_dict[md_key])

    def PoorMDFlank(self, str_dict, flank):
        self.poor_flank = poor_flank_patterns[flank].search(str_dict[flank])
        return self.poor_flank is not None

    def GrowMismatch(self, str_dict, flank):
        if self.poor_flank is None:
            raise SAMParserError("cannot grow misalign when there is no poorly "
                "aligned flank.")
        match_str = self.poor_flank.group()
        TestMsg("      Poorly aligned flank: %s" % match_str)
        if "before" == flank:
            str_dict["misalign"] = match_str + str_dict["misalign"]
            str_dict["before"] = str_dict["before"][:self.poor_flank.start()]
        elif "after" == flank:
            str_dict["misalign"] = str_dict["misalign"] + match_str
            str_dict["after"] = str_dict["after"][self.poor_flank.end():]
        else:
            raise SAMParserError("unrecognized flank: %s" % flank)
        self.poor_flank = None

    def _setup_cigar_dict(self):
        # Depends on _setup_md_dict having already been run.
        # _setup_md_dict creates three entries in the md_dict data member.
        assert 3 == len(self.md_dict)
        TestMsg("    Setting up CIGAR dict...")
        full_cigar_list = get_cigar_ops(self.raw_cigar)
        (self.cigar_qsize, self.cigar_tsize) = get_lengths_from_cigar(
            full_cigar_list)
        for md_key in self.md_dict:
            self.cigar_dict[md_key] = list()
        # TODO: Fix capitalization of member functions.
        self.CheckExistingSoftclips(full_cigar_list)
        cigar_it = iter(full_cigar_list)
        prev_cigar = leftovers = None
        # TODO: Add MISALIGN_KEYS constant.
        for key in ["before", "misalign", "after"]:
            (prev_cigar, leftovers) = self.MatchLists(self.md_dict[key], prev_cigar,
                self.cigar_dict[key], cigar_it, leftovers)
        # TODO: Determine why it is important to keep these two loops separate.
        for key in ["before", "misalign", "after"]:
            TestMsg("      %s: %s" % (key.upper(), "".join("%i%s" % (oplen,op) for
                (op,oplen) in self.cigar_dict[key])))
        # check for small flanking matches either before or after
        TestMsg("    Adjusting edges...")
        self.AdjustEdge("before", LAST)
        self.AdjustEdge("after", FIRST)
        for key in ["before", "misalign", "after"]:
            TestMsg("      %s: %s" % (key.upper(), "".join("%i%s" % (oplen,op) for
                (op,oplen) in self.cigar_dict[key])))
            TestMsg("      %s: %s" % (key.upper(), StrJoin("", self.md_dict[key])))

    def CheckExistingSoftclips(self, full_cigar_list):
        TestMsg("  Checking existing hard- or soft-clips")
        if full_cigar_list[FIRST][I_OP] in "HS":
            self.clipping[FIRST] = full_cigar_list[FIRST]
            del full_cigar_list[FIRST]
            TestMsg("    Left clip: %i%s" %
                (self.clipping[FIRST][I_LEN], self.clipping[FIRST][I_OP]))
        if full_cigar_list[LAST][I_OP] in "HS":
            self.clipping[LAST] = full_cigar_list[LAST]
            del full_cigar_list[LAST]
            TestMsg("    Right clip: %i%s" %
                (self.clipping[LAST][I_LEN], self.clipping[LAST][I_OP]))

    def AdjustEdge(self, side, index):
        if not self.cigar_dict[side]:
            TestMsg("    Nothing to adjust on %s side" % side)
            return
        TestMsg("  Adjusting %s edge" % side)
        while self.CIGARNeedsAdjustment(side, index):
            if "M" == self.cigar_dict[side][index][I_OP]:
                adjust = self.cigar_dict[side][index][I_LEN]
                old_md = int(self.md_dict[side][index])
                self.md_dict[side][index] = (old_md - adjust)
                if LAST == index:
                    self.md_dict["misalign"].insert(FIRST, adjust)
                else:
                    self.md_dict["misalign"].append(adjust)
            if LAST == index:
                self.cigar_dict["misalign"].insert(
                    FIRST, self.cigar_dict[side][index])
            else:
                self.cigar_dict["misalign"].append(self.cigar_dict[side][index])
            del self.cigar_dict[side][index]
        other_index = -1 - index
        if "N" == self.cigar_dict["misalign"][other_index][I_OP]:
            if LAST == index:
                self.cigar_dict[side].append(self.cigar_dict["misalign"][other_index])
            else:
                self.cigar_dict[side].insert(
                    FIRST, self.cigar_dict["misalign"][other_index])
            del self.cigar_dict["misalign"][other_index]

    def CIGARNeedsAdjustment(self, side, index):
        if not self.cigar_dict[side]:
            return False
        if "I" == self.cigar_dict[side][index][I_OP]:
            return True
        if ("M" == self.cigar_dict[side][index][I_OP] and
            MIN_MATCH_LENGTH > self.cigar_dict[side][index][I_LEN]):
            return True
        return False

    def MatchLists(self, md_dict, prev_cigar, curr_cigar, cigar_it,
            leftovers):
        TestMsg("  Matching lists...")
        if leftovers is not None:
            #TestMsg("    --Leftovers: %i" % leftovers)
            # if there are leftovers, split the last item added to the CIGAR list
            if 0 < leftovers and prev_cigar is not None:
                prev_cigar[LAST][I_LEN] -= leftovers
                op = prev_cigar[LAST][I_OP]
                curr_cigar.append([op, leftovers])
        #elif (None != prev_cigar):
        #  TestMsg("    --No leftovers!")
        for md_item in md_dict:
            #TestMsg("    --MD item: %s" % md_item)
            # assume the MD item is an integer (denoting matches)
            try:
                remaining = int(md_item)
            # handle non-integer MD items (denoting deletions and mismatches)
            except ValueError:
                if md_item.startswith("^"):
                    cigar_item = cigar_it.next()
                    # NOTE: CIGAR really should not have adjacent "N" and "D" operations!
                    if cigar_item[I_OP] in "IN":
                        curr_cigar.append(cigar_item)
                        cigar_item = cigar_it.next()
                    self.ProcessDelete(md_item, cigar_item, leftovers,
                        curr_cigar)
                    continue
                else:
                    remaining = len(md_item)
            #TestMsg("      Remaining: %i" % remaining)
            if leftovers is not None:
                #TestMsg("        Leftovers : %i" % leftovers)
                if leftovers > remaining:
                    #TestMsg("        Using whole MD item")
                    leftovers -= remaining
                    remaining = 0
                else:
                    #TestMsg("        Using leftovers")
                    remaining -= leftovers
                    leftovers = 0
                    #TestMsg("        Leftovers : %i" % leftovers)
                    #TestMsg("      Remaining: %i" % remaining)
            while 0 < remaining:
                (op, op_length) = cigar_it.next()
                #TestMsg("      CIGAR item: %s %i" % (op, op_length))
                # if the current operation is not an insertion to the reference or
                #  an intron, it is represented in the MD tag
                if op not in "IN":
                    if op_length > remaining:
                        leftovers = op_length - remaining
                        #TestMsg("        Creating leftovers: %i" % leftovers)
                        remaining = 0
                    else:
                        remaining -= op_length
                curr_cigar.append([op, op_length])
        return (curr_cigar, leftovers)

    # TODO: Rename ProcessDelete to _process_delete
    def ProcessDelete(self, md_item, cigar_item, leftovers, cigar_list):
        (op, op_length) = cigar_item
        TestMsg("        Cigar: %i%s" % (op_length, op))
        if (leftovers is not None and 0 < leftovers):
            raise SAMParserError("Error matching MD with CIGAR: there should not "
                "be any bases leftover from the previous CIGAR item when a "
                "deletion is encountered in the MD tag. MD:%s, CIGAR:%i%s, "
                "Leftovers:%i" % (md_item, op_length, op, leftovers))
        if "D" != op:
            raise SAMParserError("Error matching MD with CIGAR: cigar operation "
                "must be \"D\" when a deletion is encountered in the MD tag. "
                "MD:%s, CIGAR:%i%s, Leftovers:%i" %
                (md_item, op_length, op, leftovers))
        # subtract 1 from the length of the MD item (for the "^") to get the
        # length of the deletion
        del_length = len(md_item)-1
        if op_length != del_length:
            raise SAMParserError("Error matching MD with CIGAR: length of deletion "
                "in CIGAR string does not match that encountered in the MD tag. "
                "MD:%s, CIGAR:%i%s" % (md_item, op_length, op))
        cigar_list.append(cigar_item)

    def _fix_md(self):
        TestMsg('  Fixing MD')
        if md_before_and_after_should_be_merged(self.md_dict):
            merge_md_before_and_after(self.md_dict)
        self._set_fixed_md()

    def _set_fixed_md(self):
        self.md_dict['fixed'] = self.md_dict['before'] + self.md_dict['after']
        # TODO: Add detokenize_md module function?
        self.fixed_md = StrJoin('', self.md_dict['fixed'])
        TestMsg('  Fixed MD: %s' % self.fixed_md)

    def _fix_cigar(self):
        TestMsg("  Fixing CIGAR")
        # TODO: start fix_cigar_1
        (misalign_qlen, misalign_tlen) = get_lengths_from_cigar(
            self.cigar_dict['misalign'])
        # use "N" no matter what, to avoid having to add "deleted" sequence to MD tag
        # TODO: Do not need this local variable!
        removed_op = "N"
        # end fix_cigar_1
        # TODO: start fix_cigar_2
        # check for empty before or after
        # TODO: Add _misalign_is_at_front_of_cigar member function.
        if not self.cigar_dict["before"]:
            # fix_cigar_2a_misalign_start
            TestMsg("    adding \"misalign_tlen\" (%i) to tstart adjustment." %
                misalign_tlen)
            self.tadjustments[FIRST] = misalign_tlen
            if (self.cigar_dict["after"] and
                "N" == self.cigar_dict["after"][FIRST][I_OP]):
                TestMsg("      removing \"N\" at start of alignment...")
                self.tadjustments[FIRST] += self.cigar_dict["after"][FIRST][I_LEN]
                del self.cigar_dict["after"][FIRST]
        # TODO: Add _misalign_is_at_end_of_cigar member function.
        elif not self.cigar_dict["after"]:
            # fix_cigar_2b_misalign_end
            TestMsg("    adding \"misalign_tlen\" (%i) to tend adjustment." %
                misalign_tlen)
            self.tadjustments[LAST] = misalign_tlen
            if (self.cigar_dict["before"] and
                "N" == self.cigar_dict["before"][LAST][I_OP]):
                TestMsg("      removing \"N\" at end of alignment...")
                self.tadjustments[LAST] += self.cigar_dict["before"][LAST][I_LEN]
                del self.cigar_dict["before"][LAST]
        # check for "N" op before misalign
        # TODO: Add _misalign_is_after_intron.
        elif "N" == self.cigar_dict["before"][LAST][I_OP]:
            # fix_cigar_2c_misalign_after_intron
            TestMsg("    adding \"misalign_tlen\" (%i) to last \"before\" CIGAR "
                "piece." % misalign_tlen)
            self.cigar_dict["before"][LAST][I_LEN] += misalign_tlen
            if "N" == self.cigar_dict["after"][FIRST][I_OP]:
                # TODO: Is this the same as the other merge before and after?
                TestMsg("      merging before and after \"N\" ops.")
                self.cigar_dict["before"][LAST][I_LEN] += (
                    self.cigar_dict["after"][FIRST][I_LEN])
                del self.cigar_dict["after"][FIRST]
        # TODO: Add _misalign_is_before_intron.
        elif "N" == self.cigar_dict["after"][FIRST][I_OP]:
            # fix_cigar_2d_misalign_before_intron
            TestMsg("    adding \"misalign_tlen\" (%i) to first \"after\" CIGAR "
                "piece." % misalign_tlen)
            self.cigar_dict["after"][FIRST][I_LEN] += misalign_tlen
        else:
            self.cigar_dict["before"].append([removed_op, misalign_tlen])
        # end fix_cigar_2
        # TODO: start fix_cigar_3_create_fixed_list
        if not self.cigar_dict["before"]:
            self.clipping[FIRST][I_LEN] += misalign_qlen
            self.cigar_dict["fixed"] = self.cigar_dict["after"]
        elif not self.cigar_dict["after"]:
            self.clipping[LAST][I_LEN] += misalign_qlen
            self.cigar_dict["fixed"] = self.cigar_dict["before"]
        else:
            # replace with single long "I" of that size
            self.cigar_dict["fixed"] = (self.cigar_dict["before"] +
            [["I", misalign_qlen]] + self.cigar_dict["after"])
        # end fix_cigar_3_create_fixed_list
        # TODO: start fix_cigar_4_clipping
        if 0 < self.clipping[FIRST][I_LEN]:
            self.cigar_dict["fixed"].insert(FIRST, self.clipping[FIRST])
        if 0 < self.clipping[LAST][I_LEN]:
            self.cigar_dict["fixed"].append(self.clipping[LAST])
        # end fix_cigar_4_clipping
        self._set_fixed_cigar()
        TestMsg("  Fixed CIGAR: %s" % self.fixed_cigar)
        # start fix_cigar_5_check_sizes
        self.check_CIGAR_sizes()
        # end fix_cigar_5_check_sizes

    def _set_fixed_cigar(self):
        self.fixed_cigar = StrJoin("", ("%i%s" % (oplen, op) for (op, oplen) in
            self.cigar_dict["fixed"]))

    def check_CIGAR_sizes(self):
        (fixed_qsize, fixed_tsize) = get_lengths_from_cigar(self.fixed_cigar)
        if fixed_qsize != self.cigar_qsize:
            raise SAMParserError("fixing CIGAR changed qsize from %i to %i!" %
                (self.cigar_qsize, fixed_qsize))
        if fixed_tsize + sum(self.tadjustments) != self.cigar_tsize:
            raise SAMParserError("fixing CIGAR changed tsize from %i to %i + %i!" %
                (self.cigar_tsize, fixed_tsize, sum(self.tadjustments)))

    def fix_cigar_and_md_for_double_miss(self, match, is_left_match):
        helper = DoubleMissHelperCls(self.cigar_dict, self.md_dict)
        # figure out indices in 'fixed' CIGAR dict
        try:
            helper.get_match_indices(is_left_match)
        except IndexError, e:
            raise SAMParserError(
                'error in fix_cigar_and_md_for_double_miss: index error when '
                'getting fixed indices:\n{0}'.format(e))
        if 'I' != helper.get_fixed_cigar_op('insert'):
            raise SAMParserError(
                'error in fix_cigar_and_md_for_double_miss: misalign CIGAR op '
                'is not I: %s' % helper.get_fixed_cigar_op('insert'))
        # fix CIGAR
        helper.fix_cigar_for_double_miss(match)
        self._set_fixed_cigar()
        #LogMsg(self, 'Fixed CIGAR: {0}'.format(self.fixed_cigar))
        # fix MD
        helper.fix_md_for_double_miss(match)
        self._set_fixed_md()


class DoubleMissHelperCls:
    def __init__(self, cigar_dict, md_dict):
        self._cigar_dict = cigar_dict
        self._md_dict = md_dict
        self._indices = dict()
        self._last_before_index = None
        self._first_after_index = None

    def get_match_indices(self, is_left_match):
        self._get_flanking_indices()
        if is_left_match:
            self._get_left_match_indices()
        else:
            self._get_right_match_indices()

    def _get_left_match_indices(self):
        if ('M' == self._get_cigar_op('before', LAST) and
            'N' == self._get_cigar_op('after', FIRST)):
            # Element order within RNA is exon, insert, intron.
            self._indices['exon'] = self._last_before_index - 1
            self._indices['insert'] = self._last_before_index
            self._indices['intron'] = self._first_after_index
            if 2 != self._indices['intron'] - self._indices['exon']:
                raise SAMParserError(
                    'error in _get_left_match_indices: misalign portion of '
                    'CIGAR contains multiple operations.')
        elif ('M' == self._get_cigar_op('before', SECOND_TO_LAST) and
              'N' == self._get_cigar_op('before', LAST)):
            # Element order within RNA is exon, intron, insert.
            self._indices['exon'] = self._last_before_index - 2
            self._indices['intron'] = self._last_before_index - 1
            self._indices['insert'] = self._last_before_index
        else:
            raise SAMParserError(
                'error in _get_left_match_indices: cannot determine fixed '
                'indices.')

    def _get_right_match_indices(self):
        if ('M' == self._get_cigar_op('after', FIRST) and
            'N' == self._get_cigar_op('before', LAST)):
            # Element order within RNA is intron, insert, exon.
            self._indices['intron'] = self._last_before_index - 1
            self._indices['insert'] = self._last_before_index
            self._indices['exon'] = self._first_after_index
            if 2 != self._indices['exon'] - self._indices['intron']:
                raise SAMParserError(
                    'error in _get_right_match_indices: misalign portion of '
                    'CIGAR contains multiple operations.')
        elif ('M' == self._get_cigar_op('after', SECOND) and
              'N' == self._get_cigar_op('after', FIRST)):
            # Element order within RNA is insert, intron, exon.
            self._indices['insert'] = self._first_after_index - 1
            self._indices['intron'] = self._first_after_index
            self._indices['exon'] = self._first_after_index + 1
        else:
            raise SAMParserError(
                'error in _get_right_match_indices: cannot determine fixed '
                'indices.')

    def _get_flanking_indices(self):
        self._last_before_index = len(self._cigar_dict['before'])
        if self.get_fixed_cigar_op(FIRST) in 'HS':
            self._last_before_index += 1
        after_len = len(self._cigar_dict['after'])
        if self.get_fixed_cigar_op(LAST) in 'HS':
            after_len += 1
        self._first_after_index = len(self._cigar_dict['fixed']) - after_len

    def fix_cigar_for_double_miss(self, match):
        # Add match amount to exon operation, and remove it from intron and
        # insertion operations.
        self.adjust_fixed_cigar_oplen('exon', match)
        self.adjust_fixed_cigar_oplen('intron', -match)
        self.adjust_fixed_cigar_oplen('insert', -match)
        if self._match_is_too_long():
            raise SAMParserError(
                'error in fix_cigar_for_double_miss: match length {0} '
                'longer than intron {1} or insertion {2}.'.format(
                    match, self.get_fixed_cigar_oplen('intron'),
                    self.get_fixed_cigar_oplen('insert')))
        if self._intron_or_insertion_was_removed():
            self._remove_empty_ops_from_fixed_cigar()

    def _match_is_too_long(self):
        if 0 > self.get_fixed_cigar_oplen('intron'):
            return True
        if 0 > self.get_fixed_cigar_oplen('insert'):
            return True
        return False

    def _intron_or_insertion_was_removed(self):
        if (self.get_fixed_cigar_oplen('intron') and
            self.get_fixed_cigar_oplen('insert')):
            return False
        return True

    def _remove_empty_ops_from_fixed_cigar(self):
        new_fixed_list = list()
        for op, length in self._cigar_dict['fixed']:
            if length:
                new_fixed_list.append([op, length])
        self._cigar_dict['fixed'] = new_fixed_list

    def fix_md_for_double_miss(self, match):
        if self._md_is_match_before_double_miss():
            #ErrMsg('  increasing before')
            self._md_dict['before'][LAST] += match
        elif self._md_is_match_after_double_miss():
            #ErrMsg('  increasing after')
            self._md_dict['after'][FIRST] += match
        else:
            #ErrMsg('  adding new')
            self._md_dict['before'].append(match)

    def _md_is_match_before_double_miss(self):
        if (self._md_dict['before'] and
            IsInteger(self._md_dict['before'][LAST])):
            return True
        return False

    def _md_is_match_after_double_miss(self):
        if (self._md_dict['after'] and
            IsInteger(self._md_dict['after'][FIRST])):
            return True
        return False

    def _get_cigar_op(self, which_dict, dict_index):
        if isinstance(dict_index, str):
            dict_index = self._indices[dict_index]
        return self._cigar_dict[which_dict][dict_index][I_OP]

    def get_fixed_cigar_op(self, dict_index):
        return self._get_cigar_op('fixed', dict_index)

    def _get_cigar_oplen(self, which_dict, dict_index):
        if isinstance(dict_index, str):
            dict_index = self._indices[dict_index]
        return self._cigar_dict[which_dict][dict_index][I_LEN]

    def get_fixed_cigar_oplen(self, dict_index):
        return self._get_cigar_oplen('fixed', dict_index)

    def adjust_fixed_cigar_oplen(self, dict_index, amount):
        dict_index = self._indices[dict_index]
        self._cigar_dict['fixed'][dict_index][I_LEN] += amount

class LargeDeletionCls:
    def __init__(self, pattern_match, index_adjustment):
        self.start = pattern_match.start() - index_adjustment
        self.end = pattern.end() - index_adjustment
        md_token = pattern.group()
        del_size = len(md_token) - 1
        self.old_cigar = '{0}D'.format(del_size)
        self.new_cigar = '{0}N'.format(del_size)
        #TestMsg('  Found large deletion: %s\n  %s' % (
        #    self.old_cigar, md_token))
        TestMsg(
            '  Found large deletion: %s\n    length = %i, pos = %i-%i, '
            'start = %s...' % (
                self.old_cigar, len(md_token),
                self.start, self.end, md_token[:150]))

    def update_cigar(self, old_cigar):
        return old_cigar.replace(self.old_cigar, self.new_cigar)

#### EXCEPTION CLASSES ####
class SAMParserError(ContigAlignmentParserError):
  """Exception raised for errors encountered while parsing SAM format
  contig-to-genome alignments.

  """
  pass
